#!/usr/bin/perl
#
# [KSC4-PERL]---------------------------------------------------------------[]
#
#  DELTACP.PL - Copies files sending only deltas, version 0.5
#               By Marco "Kiko" Carnut <kiko@tempest.com.br> 2007-feb-19
#               Last change: 2007-apr-15
#
#  LICENSE: This program is license under the terms of the Creative Commons/
#  -------- GNU General Public License 2.0. Please see:
#
#     http://creativecommons.org/licenses/GPL/2.0/
#
#  Description: This program was mainly conceived to copy block devices from
#  ------------ one machine to the other for backup purposes, although it can
#  actually copy any file. It compares source to destination block by block
#  and transmits only those blocks that changed.
#
#  The exact difference blocks being transmitted and those that were changed
#  can be saved to other files, so as to generate forward and/or reverse
#  binary difference files that can be used to recover previous states.
#  The program also supports coalescing multiple difference files into a
#  single one.
#
#  The program must be installed and in the search path of both endpoint
#  machines. It uses SSH for channel authentication and encryption; you have
#  to make sure that you can log on to the remote account over SSH.
#
#  The program does some effort to prevent the unwary from doing dangerous
#  things like overwriting a mounted block device.
#
#  Prerequisites: Requires Perl5, preferably with LARGE_FILES support.
#  -------------- A "perl -V | grep -ic LARGE" should return 1 if your Perl
#  supports this. Most modern Perls do.
#
#  As said above, it also requires SSH.
#
#  Sample usage: Suppose /dev/vg0/lv0 is a 2GB block device and it is either a
#  ------------- snapshot with a mounted filesystem or a real logical volume
#  with an umounted filesystem. Try:
#
#     deltacp.pl /dev/vg0/lv0 user@remote.example.com:backup.img
#
#  The first the program will slowly copy everything. But if you do it again
#  right away, it will go much faster (>=10x typically) because it will detect
#  that the file didn't change. This check is not based on date and size; it
#  is hash-based in a block-by-block fashion.
#
#  If you mount the filesystem, do some stuff and unmount it again (or if
#  you release and recreate the snapshot) and run this program again, it will
#  transfer only the differences between /dev/vg0/lv0 and backup.img, so it
#  will be much less that the time to do that.
#
#  Caveats: The protocol is stop-and-wait, so it is latency-dominated. This
#  -------- makes it slow on high-latency links such as the Internet or VPNs.
#  This is not considered too bad because it makes it gentler on the network
#  and thus more suitable for background unattended operation. However, it is
#  planned that future versions of this program will address this shortcoming.
#
#  Motivation: Why not rsync, you could ask? Because as sophisticated as it
#  ----------- is, it doesn't copy the contents of block devices. If you try
#  without the --device option, it will say 'skipping non-regular file'. If
#  you specify --device, it will copy *the node*, not the contents.
#
#  Portability: Tested on Cygwin 1.5.24, Linux 2.4.29 for x86 and SPARC and 
#  ------------ Linux 2.6.17 for x86. In theory should work wherever you can
#  get a modern Perl (version 5.7.3 or greater)
#
#  WARNING: THIS PROGRAM IS IN ALPHA STAGE. It's been working for me, but
#  -------- be advised.
#
# []------------------------------------------------------------------------[]

use v5.7.3;

use strict;

use Cwd;
use Fcntl;
use POSIX;
use Symbol;
use IPC::Open3;
use Getopt::Long;
use Digest::MD5 qw(md5);

$SIG{PIPE} = \&broken_pipe;
$SIG{INT} = sub { die "Interrupted by the user"; };

# Command definitions for the communications protocol

my $cmd_false   = 0;
my $cmd_true    = 1;
my $cmd_is_hash = 2;
my $cmd_block   = 3;
my $cmd_size    = 4;

# Set up defaults

my $blksize=65536;
my $outfile;
my $logfile;
my $verbose;
my $tmpdir="/tmp/deltacp";
my $force;
my $rfile;
my $rdiff_file;
my $fdiff_file;
my $dry_run;
my $quiet;
my $xtra_args;
my $do_create;
my $no_empty_diffs;
my $do_truncate;
my $do_append;

my $start_block=0;
my $retries=50;
my $retry_pause=15;
my $max_retry_pause=300;

my $nblks_sent;
my $nblks_skipped;
my $filesize;
my $got_sigpipe;

my $prev_elapsed_time;

my $child_pid;

my %optctl=(
   "b|blocksize=i"     => \$blksize,        # Block size
   "O|outfile=s"       => \$outfile,        # Output filename
   "v|verbose+"        => \$verbose,        # Increase verbosity
   "F|force+"          => \$force,          # Proceeds even if problems found
   "d|rdiff=s"         => \$rdiff_file,     # generate reverse diff file
   "D|fdiff=s"         => \$fdiff_file,     # generate forward diff file
   "L|logfile=s"       => \$logfile,        # debug/activity _log file
   "n|dry-run+"        => \$dry_run,        # don't really write to outfile
   "x|xtra-args=s"     => \$xtra_args,      # extra arguments to ssh
   "c|create+"         => \$do_create,      # Create outfile if doesn't exist
   "t|truncate+"       => \$do_truncate,    # Try to make size(dst)=size(src)
   "e|no-empty-diffs+" => \$no_empty_diffs, # Remove diff files if empty
   "a|append+"         => \$do_append,      # Append diff files
   "q|quiet+"          => \$quiet,          # Supresses all output but errors
   "s|start-block=i"   => \$start_block,    # Block number to start from
   "r|retries=i"       => \$retries,        # How many times to retry
   "P|retry-pause=i"   => \$retry_pause     # How long to pause on retry
);

my $self_cmdline=join(" ",$0,@ARGV);

Getopt::Long::Configure("no_ignore_case","bundling");
my $result=GetOptions(%optctl);

$retry_pause=15 if $retry_pause<1;
$retry_pause=$max_retry_pause if $retry_pause>$max_retry_pause;

my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);

$|=1; # Turn on autoflush
my $fd;
my $cblock=$start_block;

if ($outfile) {  # ---------- Patch ("receiver") mode

    my %already;
    my %offset;
    
    my $irdiff_file;
    my $ifdiff_file;
    
    my $filesize_msg;

    my $isodate=sprintf("%02d%02d%02d",$year%100,$mon+1,$mday);
    my ($outfname) = $outfile =~ m|/?([^/]+?)$|o;
    
    $rdiff_file="rdiff-$outfname-$isodate" if $rdiff_file eq "-";
    $fdiff_file="fdiff-$outfname-$isodate" if $fdiff_file eq "-";

    # Perform file in use detection, which starts with finding out what
    # exactly the local file is. For this end, we first resolve symlinks
    # and canonicalize it.
    $outfile=expand_file_name(dereference_links($outfile));
        
    # Return an error if the file is already in use by another instance of
    # this program. The SSH channel will propagate this error code back to the
    # sender process. It will get a SIGPIPE, which is trapped by the END block
    # There a waitpid call retrieves the return code and displays a nice
    # error message.
    if (check_in_use($outfile,$force>1) && !$force) {
        if (!$quiet) {
            print STDERR "FATAL: looks like '$outfile' is in use by ";
            print STDERR "another instance of deltacp.\n";
        }
        exit 100;
    }
    
    # Return an error if the file is mounted. The error code will be
    # propagated using the same processe described above. Notice that
    # detecting loopbacked mounts is not reliable.
    if (is_mounted($outfile) && !$force) {
        if (!$quiet) {
            print STDERR "FATAL: looks like '$outfile' is mounted\n";
        }
        exit 101;
    }

    my $mode=$dry_run ? &POSIX::O_RDONLY : &POSIX::O_RDWR;
    $mode |= &POSIX::O_CREAT if $do_create;
    if (!($fd=POSIX::open($outfile,$mode))) {
        if (!$quiet) {
            print STDERR "FATAL: Unable to open '$outfile' for random-acess ";
            print STDERR ($dry_run ? "reading\n" : "read/write\n");
        }
        # Return an error if we can't open $outfile properly. This value will
        # be propagated back by the same mechanism described above.
        exit 102;
    }
    mark_in_use($outfile);
    
    binmode STDIN;
    binmode STDOUT;
    
    open(LOGFH,">$logfile") && close(LOGFH) if $logfile;
    if ($rdiff_file) {
        if ($rdiff_file !~ m|^/|o) {
            if ($outfile =~ m|^/dev/|o) {
                $rdiff_file = expand_file_name(dereference_links($rdiff_file));
                my ($basedir,$fname) = $rdiff_file =~ m|(.*)/(.*)|o;
                $irdiff_file = "$basedir/.$fname";                
            } else {
                my ($basedir) = $outfile =~ m|(.*)/|o;
                $irdiff_file = "$basedir/.$rdiff_file";
                $rdiff_file  = "$basedir/$rdiff_file";
            }
        } else {
            my ($basedir,$fname) = $rdiff_file =~ m|(.*)/(.*)|o;
            $irdiff_file = "$basedir/.$fname";
        }
        if (open(RDFH,$rdiff_file)) {
            close(RDFH);
            if ($do_append) {
                rename($rdiff_file,$irdiff_file);
            } else {
                if ($force) {
                    unlink($rdiff_file);
                } else {
                    print STDERR "Won't cobbler existing reverse diff file ".
                                 "'$rdiff_file' without --force.\n";
                    exit 105;
                }
            }
        }
        # Is there an incomplete diff file with this name? If so, let's
        # use it and restore it's 'already' status
        if (open(RDFH,"+<$irdiff_file")) {
            binmode RDFH;
            for (;;) {
                my $msg;
                my $nbytes=read(RDFH,$msg,25);
                last if $nbytes<25;
                my ($cmd,$cblock,$length,$hash) = unpack("cN2a16",$msg);
                if ($cmd == $cmd_block) {
                    $already{$cblock}=1;
                }
                seek(RDFH,$length,SEEK_CUR);
            }
            seek(RDFH,0,SEEK_END);
        } else {
            if (!open(RDFH,">$irdiff_file")) {
                print STDERR "Unable to open '$rdiff_file' for writing.\n";
                exit 103;
            }
            binmode RDFH;
        }
    }
   if ($fdiff_file) {
        if ($fdiff_file !~ m|^/|o) {
            if ($outfile =~ m|^/dev/|o) {
                $fdiff_file = expand_file_name(dereference_links($fdiff_file));
                my ($basedir,$fname) = $fdiff_file =~ m|(.*)/(.*)|o;
                $ifdiff_file = "$basedir/.$fname";                
            } else {
                my ($basedir) = $outfile =~ m|(.*)/|o;
                $ifdiff_file = "$basedir/.$fdiff_file";
                $fdiff_file  = "$basedir/$fdiff_file";
            }
        } else {
            my ($basedir,$fname) = $fdiff_file =~ m|(.*)/(.*)|o;
            $ifdiff_file = "$basedir/.$fname";
        }
        if (open(FDFH,$fdiff_file)) {
            close(FDFH);
            if ($do_append) {
                rename($fdiff_file,$ifdiff_file);
            } else {
                if ($force) {
                    unlink($fdiff_file);
                } else {
                    print STDERR "Won't cobbler existing forward diff file ".
                                 "'$fdiff_file' without --force.\n";
                    exit 106;
                }
            }
        }
        if (open(FDFH,"+<$irdiff_file")) {
            binmode FDFH;
            for (;;) {
                my $msg;
                my $nbytes=read(FDFH,$msg,25);
                last if $nbytes<25;
                my ($cmd,$cblock,$length,$hash) = unpack("cN2a16",$msg);
                if ($cmd == $cmd_block) {
                    $offset{$cblock}=tell(FDFH);
                }
                seek(FDFH,$length,SEEK_CUR);
            }
            seek(FDFH,0,SEEK_END);
        } else {
            if (!open(FDFH,"+>$ifdiff_file")) {
                print STDERR "Unable to open '$fdiff_file' for writing.\n";
                exit 104;
            }
            binmode FDFH;
        }
    }
    
    seek(STDIN,0,SEEK_END);
    my $stdin_size=tell(STDIN);
    seek(STDIN,0,SEEK_SET);
    
    for (;;) {
        my ($msg,$lbuf,$buf);
        _log(" wait: waiting for 25-byte message\n");
        my $nbytes=read(STDIN,$msg,25);
        if ($stdin_size) {
            my $stdin_pos=tell(STDIN);
            printf(STDERR "%5.1f%%\r",$stdin_pos*100.0/$stdin_size);
        }
        last if $nbytes<25;
        _log(" recv: ".unpack("H*",$msg)."\n");
        my ($cmd,$cblock,$length,$hash) = unpack("cN2a16",$msg);
        _log("       decoded as $cmd: $cblock,$length\n");
        if ($cmd == $cmd_size) {
            $filesize = $cblock*$blksize+$length;
            $filesize_msg = $msg;
            next;
        }
        if ($cmd != $cmd_is_hash && $cmd != $cmd_block) {
            _log("error: invalid command\n");
            exit 199;
        }
        POSIX::lseek($fd,$cblock*$blksize,&POSIX::SEEK_SET);
        POSIX::read($fd,$lbuf,$blksize);
        my $lhash=md5($lbuf);
        _log("lhash: ".unpack("H*",$lhash)."\n");
        my $reply=$lhash ne $hash ? $cmd_true : $cmd_false;
        if ($cmd == $cmd_block) {
            $reply=$cmd_true;
        } else {
            print chr($reply);
            _log(" sent: CMD_".($reply?"TRUE":"FALSE")."\n");
        }
        if ($reply) {
            $nbytes=read(STDIN,$buf,$length);
            if ($nbytes<0) {
                _log("error: Unexpected end of stream");
                last;
            }
            _log(" recv: $nbytes data block\n");
            if ($nbytes>0) {
                if ($rdiff_file) {
                    if (!$already{$cblock}) {
                        print RDFH pack("cN2a16",$cmd_block,$cblock,$length,
                                        $lhash);
                        print RDFH $lbuf;
                        $already{$cblock}=1;
                    }
                }
                my $rhash=md5($buf);
                if ($rhash eq $hash) {
                    if ($fdiff_file) {
                        my $seekto=$offset{$cblock};
                        if ($seekto eq undef) {
                            seek(FDFH,0,SEEK_END);
                            $offset{$cblock}=tell(FDFH);
                        } else {
                            seek(FDFH,$seekto,SEEK_SET);
                        }
                        print FDFH pack("cN2a16",$cmd_block,$cblock,$length,
                                        $hash);
                        print FDFH $buf;
                    }
                    if (!$dry_run) {
                        POSIX::lseek($fd,$cblock*$blksize,&POSIX::SEEK_SET);
                        my $nwbytes=POSIX::write($fd,$buf,$nbytes);
                        _log("wrote: block $cblock, nwb=$nwbytes, errno=$!\n");
                    }
                } else {
                    _log(" skip: bad hash on block $cblock\n");
                }
            }
        }
    }    
    
    POSIX::close($fd) if !$dry_run;
    
    if ($do_truncate && $filesize && ! -b $outfile) {
        if (open(TFH,"+<$outfile")) {
            truncate(TFH,$filesize);
            close(TFH);
        }
    }
    
    if ($rdiff_file) {
        my $size=tell(RDFH);
        if ($no_empty_diffs && !$size) {
            close(RDFH);
            unlink($irdiff_file);
        } else {
            print RDFH $filesize_msg if $filesize_msg;
            close(RDFH);
            rename($irdiff_file,$rdiff_file);
        }
    }
    if ($fdiff_file) {
        my $size=tell(FDFH);
        if ($no_empty_diffs && !$size) {
            close(FDFH);
            unlink($ifdiff_file);
        } else {
            print FDFH $filesize_msg if $filesize_msg;
            close(FDFH);
            rename($ifdiff_file,$fdiff_file);
        }
    }
    
} else {         # ---------- Normal ("diff generation") mode

    my $srcdev=shift;
    my $dest=shift;
    
    if (!$srcdev || !$dest) {
        usage(); exit 0;
    }
    
    if ($dest =~ /\@:/o) {
        print "Invalid destination syntax: if you specify a user, you've ";
        print "got to specify a host as well\n";
        exit 1;
    }

    # Parse [[user@]host:]file and provide appropriate defaults
    my ($user,$host,$file);
    $dest .= ":" if $dest =~ /@/o && $dest !~ /:/o;
    (undef,$user,undef,$host,$file) = $dest=~/((\w+)@)?(([\w.]+):)?(.*?)?$/o;
    $file = $srcdev if !$file;
    $user = $ENV{USER} || $ENV{USERNAME} if !$user;
    $rfile = $file;
    print "user=$user, host=$host, file=$file\n" if $verbose>1;

    # Perform local file in use detection, which starts with finding out
    # what exactly the local file is. For this end, we first resolve symlinks
    # and canonicalize it
    $srcdev=expand_file_name(dereference_links($srcdev));
    print "resolved canonicalized source file: $srcdev\n" if $verbose;
    
    if (is_mounted($srcdev)) {
        if ($force) {
            print "Warning: '$srcdev' is mounted.\n";
        } else {
            print "FATAL: '$srcdev' is mounted.\n";
            exit 1;
        }
    }
    
    if (my $pid=check_in_use($srcdev)) {
        if ($force) {
            print "Warning: it appears that \"$srcdev\"\n";
            print "  is in use by another deltacp instance with pid $pid.\n";
            print "  However, you requested --force, so I'll proceed. I\n";
            print "  hope that you are aware that two instances of deltacp\n";
            print "  operating on the same file at the same time is almost\n";
            print "  a guarantee that they will end up corrupted and for\n";
            print "  some reason this is what you really want.\n";
        } else {
            print "FATAL: it appears that \"$srcdev\"\n";
            print "  is in use by another deltacp instance with pid $pid.\n";
            print "  As two instances of deltacp working on the same file\n";
            print "  at the same time is almost a guarantee that the copy\n";
            print "  will be corrupted, I'll cowardly refuse to proceed.\n";
            print "  Do a \'kill $pid\' if you want to proceed safely.\n\n";
            print "  If you absolutely know that you're doing, you can skip\n";
            print "  this check with --force.\n";
            exit 1;
        }
    }

    mark_in_use($srcdev);
        
    my $INFH=gensym;
    my $OUTFH=gensym;
    my $ERRFH=gensym;
    my $cmdline;
    if ($user && $host) {
        $cmdline.="ssh $user\@$host ";
        $cmdline.="$xtra_args " if $xtra_args;
    }
    $cmdline.="'deltacp.pl -qc -O $file -b $blksize";
    $cmdline.=" -n" if $dry_run;
    $cmdline.=" -F" if $force;
    $cmdline.=" -a" if $do_append;
    $cmdline.=" -e" if $no_empty_diffs;
    $cmdline.=" -L $logfile" if $logfile;
    $cmdline.=" -t " if $do_truncate;
    $cmdline.=" -d $rdiff_file " if $rdiff_file;
    $cmdline.=" -D $fdiff_file " if $fdiff_file;
    $cmdline.=" 2>/dev/null'";
    print "cmdline: $cmdline\n" if $verbose;
    
    $got_sigpipe=0;
    
    my ($ete,$etr);
    my $blocks_per_second;
    my $start_time;
    my ($last_block,$last_length);
    if ($child_pid=open3($OUTFH,$INFH,$ERRFH,$cmdline)) {
        binmode $INFH;
        binmode $OUTFH;
        
        if (!sysopen(FH,$srcdev,O_RDONLY)) {
            print "Can't open $srcdev for reading";
            exit 1;
        }
        my $size=sysseek(FH,0,SEEK_END);
        my $nblocks=int(($size+$blksize-1)/$blksize);
        die "Illegal starting block" if $cblock<0 || $cblock>$nblocks;
        sysseek(FH,$cblock*$blksize,SEEK_SET);
                
        for (;;) {
            my $buf;
            my $length=sysread(FH,$buf,$blksize);
            last if $length<=0 || $got_sigpipe; 
            my $hash=md5($buf);
            my $msg=pack("cN2a16",$cmd_is_hash,$cblock,$length,$hash);
            $last_block=$cblock;
            $last_length=$length;
            print $OUTFH $msg;
            if ($verbose>1) {
                print "sent: IS_HASH($cblock,$length)\n";
                print "      ",unpack("H*",$msg),"\n";
            }
            my $reply;
            my $nbytes=read($INFH,$reply,1);            
            last if $got_sigpipe;
            if ($nbytes<1) {
                print "fatal: Unexpected end of stream\n"; 
                print <$ERRFH>;
                $got_sigpipe=1;
                last;
            }
            print "recv: got reply ",ord($reply),"\n" if $verbose>1;
            if (ord($reply)==$cmd_true) {
                print "recv: CMD_TRUE\nsent: block $cblock, length $length\n"
                       if $verbose>1;
                print $OUTFH $buf;
                last if $got_sigpipe;
                $nblks_sent++;
            } else {
                print "recv: CMD_FALSE\n" if $verbose>1;
                $nblks_skipped++;
            }
            $cblock++;
            if (!$quiet) {
                if (!$quiet && $cblock==1) {
                    print "  current/total   current% |";
                    print " skips%      sent/skips\n";
                }
                printf("%9d/",$cblock);
                if ($size) {
                    printf("%-9d ",$nblocks);
                    printf("%5.1f%% ",int($cblock*1000.0/$nblocks)/10.0);
                } else {
                    print "--------- ---.-% "
                }
                printf("| %5.1f%%",int($nblks_skipped*1000.0/$cblock)/10.0);
                printf(" %9d/%-9d ",$nblks_sent,$nblks_skipped);
    
                if ($start_time>0 && $size) {    
                    my $elapsed_time=time()-$start_time;
                    if ($elapsed_time>5) {
                        if ($elapsed_time!=$prev_elapsed_time) {
                            $blocks_per_second=$cblock/$elapsed_time;
                            $ete=$nblocks/$blocks_per_second;
                            $etr=$ete-$elapsed_time;
                            my $sec=$etr%60; $etr/=60;
                            my $min=$etr%60; $etr/=60;
                            my $hrs=$etr%24;
                            my $day=int($etr/24);
                            printf("%3dd %02d:%02d:%02d ",$day,$hrs,$min,$sec);
                            $prev_elapsed_time=$elapsed_time;
                        }
                    } else {
                        $blocks_per_second=0;
                        $ete=$etr=0;
                    }
                } else {
                    $start_time=time();
                    print "---d --:--:--";
                }
                print $verbose?"\n":"\r";
            }
        }
        my $hash=chr(0) x 16;
        my $msg=pack("cN2a16",$cmd_size,$last_block,$last_length,$hash);
        print $OUTFH $msg;
        close($OUTFH);
        close($INFH);
        close($ERRFH);
    }
    close(FH);
    if ($got_sigpipe) {
        print STDERR "\nGot SIGPIPE, pausing for $retry_pause seconds\n";
        sleep($retry_pause);
        $self_cmdline =~ s/(-x|--xtra_args)(=|\s+)$xtra_args//;
        $self_cmdline =~ s/(-r|--retries)(=|\s+)\d+//o;
        $self_cmdline =~ s/(-P|--retry-pause)(=|\s+)\d+//o;
        $self_cmdline =~ s/(-s|--start_block)(=|\s+)\d+//o;
        $self_cmdline =~ s/-FF//o;
        $self_cmdline =~ s/\s{2,}/ /o;
        $retry_pause*=2;
        $retry_pause=$max_retry_pause if $retry_pause>$max_retry_pause;
        $retries--;
        if ($retries) {
            print STDERR "Retrying ($retries left)...\n";
            remove_in_use_mark();
            $cblock-- if $cblock>0; 
            $self_cmdline .= " -r $retries -P $retry_pause -s $cblock -FF";
            $self_cmdline .= " -a" if $rdiff_file || $fdiff_file;
            my @cmdline=split(/\s+/o,$self_cmdline);
            push @cmdline,"-x",$xtra_args if $xtra_args;
            if ($verbose) {
                print STDERR "new cmdline: ",join(" ",@cmdline),"\n";
            }
            exec @cmdline;
        } else {
            print STDERR "Maximum retry count exceeded, giving up.\n";
        }
    }
    
    print "\n" if !$verbose && !$quiet;
}

sub _log
{
    return if !$logfile;
    my ($msg) = @_;
    if (open(LOGFH,">>$logfile")) {
        print LOGFH $msg;
        close(LOGFH);
    }
}

sub check_in_use
{
    my ($srcfile,$do_kill) = @_;
    opendir(PFD,$tmpdir) || return 0;
    while (my $fname=readdir(PFD)) {
        next if $fname !~ /^\d/;
        if (open(PFH,"$tmpdir/$fname")) {
            my $file=<PFH>;
            close(PFH);
            if ($srcfile eq $file) {
                # Check if the process with this PID is still alive and
                # it is in fact a deltacp instance
                if (open(UFH,"/proc/$fname/cmdline")) {
                    my $text=<UFH>;
                    close(UFH);
                    if ($text =~ /deltacp/oi) {
                        if ($do_kill) {
                            kill 15,$fname;
                        } else {
                            return $fname;
                        }
                    }
                }
            }
        }
    }
    sleep(2) if $do_kill;
    close(PFD);
    return 0;
}

sub mark_in_use
{
    my ($file) = @_;
    if (! -d $tmpdir ) {
        mkdir($tmpdir);
        chmod(0777,$tmpdir);
    }
    my $tmpfname="$tmpdir/$$";
    if (open(UFH,">$tmpfname")) {
        print UFH $file;
        close(UFH);
    } else {
        print "Warning: Unable to create $tmpfname to mark file as in use.\n";
        print "  Detection of multiple instances working on the same file\n";
        print "  will almost certainly fail.\n";
    }
}

sub remove_in_use_mark
{
    unlink("$tmpdir/$$");
}

END
{
    print "removing in-use mark\n" if $verbose;
    remove_in_use_mark();
}

# FIXME: This routine more or less assume that the SIGPIPE will only come from
# the SSH subprocess, but it can also come from the input file when it is,
# well, a pipe. :) Currently the routine does not handle that.

sub broken_pipe
{
    if ($child_pid) {
        waitpid($child_pid,0);
        my $errcode=$?>>8;
        if ($errcode==100) {
            print "FATAL: The *remote* file is in use by another deltacp ";
            print "instance. Given that two\n  deltacp instances writing ";
            print "to the same file at the same almost invariably\n  results";
            print " in your file becoming corrputed, I will not proceeed.";
            print "If you like,\n  log on the destination machine, kill ";
            print "the other deltacp instances operating\n  in that file ";
            print "and try again.\n";
            exit $errcode;
        } elsif ($errcode==101) {
            print "FATAL: Looks like the *remote* file '$rfile' is mounted.\n";
            exit $errcode;
        } elsif ($errcode==102) {
            print "FATAL: The *remote* process was unable to open '$rfile'.";
            print "\n  Check that the path really exists and that you have";
            print " appropriate permissions\n  there.\n";
            exit $errcode;
        } elsif ($errcode==103) {
            print "FATAL: The remote process was unable to open the reverse ";
            print "diff file\n  '$rdiff_file' for writing.\n";
            exit $errcode;
        } elsif ($errcode==104) {
            print "FATAL: The remote process was unable to open the forward ";
            print "diff file\n  '$fdiff_file' for writing.\n";
            exit $errcode;
        } elsif ($errcode==105) {
            print "FATAL: The remote process reports that the reverse ";
            print "diff file\n  '$rdiff_file' already exists.\n  We're not ";
            print "sure that overwriting it is what you really want.\n  ";
            print "If it is, use --force\n";
            exit $errcode;
        } elsif ($errcode==106) {
            print "FATAL: The remote process reports that the forward ";
            print "diff file\n  '$fdiff_file' already exists.\n  We're not ";
            print "sure that overwriting it is what you really want.\n  ";
            print "If it is, use --force\n";
            exit $errcode;
        } else {
            $got_sigpipe=1;
        }
    }
}

sub is_mounted
{
    return 0 if $^O =~ /cygwin/oi;
    my ($file) = @_;
    my $rfile=$file;
    $rfile =~ s|.*/||o;
    for (`mount`) {
        my ($mount_point) = split(/\s+/o);
        next unless $mount_point =~ m|^/|o;
        $mount_point=expand_file_name(dereference_links($mount_point));
        return 1 if $mount_point eq $file;
        my $text = `losetup $mount_point 2>&1`;
        my ($loop_file) = $text =~ /$mount_point:.*?\((.*?)\)$/;
        $loop_file =~ s|.*/||o;
        return 1 if $loop_file eq $rfile;
    }
    return 0;
}

sub usage
{
print <<EOF;
deltacp v0.5 - Copy files transmitting only differences
               By Marco "Kiko" Carnut <kiko at tempest dot com dot br>
               2007-apr-15
               
Usage: deltacp [options] <source> [[<user>@]<host>:]<destination>

Options are:
-v        verbose/debug, repeat for more
-q        quiet, supress all output except errors
-b #      specify block size in bytes (default is 65536)
-O file   patch file, diffs come in the standard input
-d file   generate reverse diff (use "-" to autogenerate the file name)
-D file   generate forward diff (use "-" to autogenerate the file name)
-L file   ask remote to generate debug logfile
-F        force, supress safety checks (twice: harder, kills remotes)
-n        dry run, does it all but doesn't really write to the dest file
-c        create output file if it doesn't exist
-t        truncate: make output file the same length as source file
-e        remove diff files if they turn out to be empty
-a        append to diff files
-s #      start block
-r #      retry count, default zero
-p #      pause this many secs between retries, default 15
EOF
}

# --------------------------- Auxiliary Routines ----------------------------

# The two routines below are used elsewhere to convert symbolic links to
# absolute canonicalized names of the files links point to. They've been
# written by Noah Friedman <friedman@splode.com>. See:
# http://www.splode.com/~friedman/software/scripts/src/symlink-resolve

# Args: basename, default-directory
#
# Convert basename to absolute, and canonicalize it.  Second arg
# default-directory is directory to start with if basename is relative
# (does not start with slash); if default-directory is undefined, the
# current working directory is used.  File name components that are `.' are
# removed, and so are file name components followed by `..', along with the
# `..' itself; note that these simplifications are done without checking
# that the file name actually exists in the file system.
sub expand_file_name ($;$)
{
  local $_ = shift;

  my $cwd;
  if ($_ !~ m|^/|o)
    {
      $cwd = getcwd() unless (defined $cwd);
      $_ = join ('/', (shift || $cwd), $_);
    }

  # These substitutions must be done in loops to handle overlapping `/'
  # characters in adjacent patterns.
  s|/\./|/|o         while (m|/\./|o);
  s|//|/|o           while (m|//|o);
  s|/[^/]+/\.\./|/|o while (m|/[^/]+/\.\./|o);
  s|/[^/]+/\.\.$||o;
  s|/.$||go;
  # Eliminate leading `..'.
  # It may be harmful to do it if the filesystem interprets `/..' as
  # something not equivalent to `/'.
  #s|^/\.\./|/|o     while (m|^/\.\./|o);
  return $_;
}

sub dereference_links ($)
{
  my $file = shift;
  my @p = split (m|/|, $file);
  my $link_count = 0;
  for (my $j = 0; $j <= $#p; $j++)
    {
      my $k = join ("/", @p[0 .. $j]);
      my $orig_component = $k;
      while (my $l = readlink ($k))
        {
          $k = $l;
          # Simple way of detecting symlink loops (it unfortunately causes
          # the system to give up when there are simply too many levels,
          # even if resolution would eventually occur).  This parameter is
          # adjustable, of course.  Most unix kernels allow a depth of 8.
          return undef if $link_count++ == 64;
        }
      next if ($k eq $orig_component);
      if (substr ($k, 0, 1) eq "/")
        {
          # Absolute link.  Trash $p[0]-$p[$j+1] and replace with readlinked
          # path components.  Set $j to -1 so that next iteration of loop
          # will check array @p from start.
          splice (@p, 0, $j + 1, split (m|/|, $k));
          $j = -1;
        }
      else
        {
          # Insert partial (relative) path component into array in place of
          # current element $p[$j]
          splice (@p, $j, 1, split (m|/|, $k));
          $j--;
        }
    }
  join ("/", @p);
}
